#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <getopt.h>

#include "config.h"
#include "log.h"
#include "service.h"
#include "services.h"
#include "util.h"

#define CFG_DEFAULT_FILENAME "/etc/yaddns.conf"

#define CFG_DEFAULT_WANIFNAME "ppp0"

/* #define CFG_DEFAULT_MYIP_HOST "checkip.dyndns.org" */
/* #define CFG_DEFAULT_MYIP_PORT 80 */
/* #define CFG_DEFAULT_MYIP_PATH "/" */
/* #define CFG_DEFAULT_MYIP_UPINT 60 */

#define CFG_FREE(itname) do {                                           \
                if(itname != NULL)                                      \
                {                                                       \
                        free(itname);                                   \
                        itname = NULL;                                  \
                }                                                       \
        } while(0)

#define CFG_SET_VALUE(itname, value) do {                               \
                CFG_FREE(itname);                                       \
                itname = strdup(value);                                 \
        } while(0)

/*
 * spaces = space, \f, \n, \r, \t and \v
 */
static int config_get_assignment(FILE *file, char *buffer, size_t buffer_size,
                                 const char *filename, int *linenum,
                                 char **name, char **value)
{
        char *t = NULL, *end = NULL;
        char *n = NULL, *equals = NULL, *v = NULL;
        int ret = -1;

        while(fgets(buffer, buffer_size, file) != NULL)
	{
                (*linenum)++;

                /* search \n */
                end = strchr(buffer, '\n');
                if(!end)
                {
                        /* case of last line not terminated by \n */
                        end = buffer + strlen(buffer) - 1;
                }

                /* remove spaces at end */
                while(end > buffer && isspace(*end))
                {
                        *end = '\0';
                        --end;
                }

                /* remove spaces at start */
                n = buffer;
                while(isspace(*n))
                {
                        n++;
                }

                /* skip comment or empty lines */
		if(n[0] == '#' || n[0] == '\0')
                {
                        continue;
                }

                /* account definition ? */
                if(memcmp(n, "account", sizeof("account") - 1) == 0)
                {
                        /* maybe a account line definition ? */
                        if((equals = strchr(n, '{')) != NULL)
                        {
                                /* remove whitespaces before { */
                                for(t = equals - 1;
                                    t > n && isspace(*t);
                                    t--)
                                {
                                        *t = '\0';
                                }

                                /* maybe { is glued to account ? */
                                *equals = '\0';

                                *name = n;
                                *value = NULL;
                                ret = 0;
                        }
                        else
                        {
                                log_error("parsing error at '%s'. Invalid "
                                          "account declaration. "
                                          "(file %s - line %d)",
                                          n, filename, (*linenum));
                                ret = -1;
                        }

                        break;
                }

                /* maybe end of account definition ? */
                if(memcmp(n, "}", sizeof("}") - 1) == 0)
                {
                        *name = NULL;
                        *value = NULL;
                        ret = 0;
                        break;
                }

                /* need a '=' for a correct assignement  */
                if(!(equals = strchr(n, '=')))
		{
			log_error("parsing error at '%s'. Invalid assignment. "
                                  "(file %s - line %d)",
                                  n, filename, (*linenum));
			break;
		}

                /* remove whitespaces before = */
                for(t = equals - 1;
                    t > n && isspace(*t);
                    t--)
                {
                        *t = '\0';
                }

                /* maybe = is glued to "name" ? */
                *equals = '\0';

		v = equals + 1;

                /* skip spaces */
		while(isspace(*v))
                {
			v++;
                }

                /* remove " if present */
                if(*v == '"')
                {
                        v++;
                }

                /* remove " at end if present */
                if(*end == '"')
                {
                        *end = '\0';
                        --end;
                }

                *name = n;
                *value = v;
                ret = 0;

                break;
        }

        return ret;
}

int config_parse(struct cfg *cfg, int argc, char **argv)
{
        int cfgfile_flag = 0;
        struct service *service = NULL;

        int c, ind;
        char short_options[] = "vhlf:p:D";
        struct option long_options [] = {
                {"version", no_argument, 0, 'v' },
                {"help", no_argument, 0, 'h' },
                {"list-service", no_argument, 0, 'l' },
                {"daemonize", no_argument, 0, 'D' },
                {"cfg", required_argument, 0, 'f' },
                {"pid-file", required_argument, 0, 'p' },
                {"daemonize", required_argument, 0, 'D' },
                {0, 0, 0, 0 }
        };

        /* void cfg variables which can be set in cmd line */
        cfg->cfgfile = NULL;
        cfg->pidfile = NULL;
        cfg->daemonize = 0;

        while((c = getopt_long (argc, argv,
                                short_options, long_options, &ind)) != EOF)
        {
                switch(c)
                {
                case 0:
                        break;

                case 'h':
                        printf("Usage: %s [-f cfg_filename] [-p pid_file] [-D]\n", argv[0]);
                        printf("Options:\n");
                        printf(" -f, --cfg=FILE\t\tConfig file to be used\n");
                        printf(" -p, --pid-file=FILE\tPID file to be used\n");
                        printf(" -D, --daemonize\t\tDaemonize yaddns\n");
                        printf(" -h, --help\t\tDisplay this help\n");
                        printf(" -l, --list-service\tDisplay the "
                               "list of available services\n");
                        printf(" -v, --version\t\tDisplay the version\n");
                        return -1;

                case 'v':
                        printf(PACKAGE_STRING "\n");
                        return -1;

                case 'l':
                        printf("List of available services:\n");
                        list_for_each_entry(service,
                                            &service_list, list)
                        {
                                printf("  - %s\n", service->name);
                        }
                        return -1;

                case 'D':
                        cfg->daemonize = 1;
                        break;

                case 'p':
                        cfg->pidfile = strdup(optarg);
                        break;

                case 'f':
                        cfg->cfgfile = strdup(optarg);
                        cfgfile_flag = 1;
                        break;

                default:
                        break;
                }
        }

        if(cfg->cfgfile == NULL)
        {
                cfg->cfgfile = strdup(CFG_DEFAULT_FILENAME);
        }

        if(config_parse_file(cfg, cfg->cfgfile) != 0)
        {
                if(access(cfg->cfgfile, F_OK) == 0 || cfgfile_flag)
                {
                        config_free(cfg);
                        return -1;
                }
        }

        /* check if there account defined */
        if(list_empty(&(cfg->accountcfg_list)))
        {
                log_warning("No account defined.");
        }

	return 0;
}

int config_parse_file(struct cfg *cfg, const char *filename)
{
	FILE *file = NULL;
        int ret = 0;
        long n = 0;
	char buffer[1024];
	int linenum = 0;
	char *name = NULL, *value = NULL;
        int accountdef_scope = 0;
        struct accountcfg *accountcfg = NULL,
                *safe_accountcfg = NULL;
        int myip_assign_count = 0;


        /* void cfg variables which can be set in cfg file */
        cfg->wan_cnt_type = wan_cnt_direct;
        cfg->wan_ifname = NULL;
        cfg->myip.host = NULL;
        cfg->myip.port = 0;
        cfg->myip.path = NULL;
        cfg->myip.upint = 0;

        log_debug("Trying to parse '%s' config file", filename);

        if((file = fopen(filename, "r")) == NULL)
        {
                log_error("Error when trying to open '%s' config file: %m",
                          filename);
		return -1;
        }

        while((ret = config_get_assignment(file, buffer, sizeof(buffer),
                                           filename, &linenum,
                                           &name, &value)) != -1)
	{
                log_debug("assignment '%s' = '%s'", name, value);

                if(accountdef_scope)
                {
                        if(name == NULL)
                        {
                                accountdef_scope = 0;

                                /* check and insert */
                                if(accountcfg->name == NULL
                                   || accountcfg->service == NULL
                                   || accountcfg->username == NULL
                                   || accountcfg->passwd == NULL
                                   || accountcfg->hostname == NULL)
                                {
                                        log_error("Missing value(s) for "
                                                  "account name '%s' service '%s'"
                                                  "(file %s - line %d)",
                                                  accountcfg->name,
                                                  accountcfg->service,
                                                  filename, linenum);

                                        CFG_FREE(accountcfg->name);
                                        CFG_FREE(accountcfg->service);
                                        CFG_FREE(accountcfg->username);
                                        CFG_FREE(accountcfg->passwd);
                                        CFG_FREE(accountcfg->name);
                                        free(accountcfg);

                                        ret = -1;
                                        break;
                                }

                                list_add(&(accountcfg->list),
                                         &(cfg->accountcfg_list));
                        }
                        else if(strcmp(name, "name") == 0)
                        {
                                CFG_SET_VALUE(accountcfg->name, value);
                        }
                        else if(strcmp(name, "service") == 0)
                        {
                                CFG_SET_VALUE(accountcfg->service, value);
                        }
                        else if(strcmp(name, "username") == 0)
                        {
                                CFG_SET_VALUE(accountcfg->username, value);
                        }
                        else if(strcmp(name, "password") == 0)
                        {
                                CFG_SET_VALUE(accountcfg->passwd, value);
                        }
                        else if(strcmp(name, "hostname") == 0)
                        {
                                CFG_SET_VALUE(accountcfg->hostname, value);
                        }
                        else
                        {
                                log_error("Invalid option name '%s' for "
                                          "an account (file %s line %d)",
                                          name, filename, linenum);

                                ret = -1;
                                break;
                        }
                }
                else if(strcmp(name, "account") == 0)
                {
                        accountdef_scope = 1;
                        accountcfg = calloc(1, sizeof(struct accountcfg));
                        log_debug("add accountcfg '%p'", accountcfg);
                }
                else if(strcmp(name, "wanifname") == 0)
                {
                        CFG_SET_VALUE(cfg->wan_ifname, value);
                }
                else if(strcmp(name, "mode") == 0)
                {
                        if(strcmp(value, "indirect") == 0)
                        {
                                cfg->wan_cnt_type = wan_cnt_indirect;
                        }
                        else
                        {
                                cfg->wan_cnt_type = wan_cnt_direct;
                        }
                }
                else if(strcmp(name, "myip_host") == 0)
                {
                        CFG_SET_VALUE(cfg->myip.host, value);
                        ++myip_assign_count;
                }
                else if(strcmp(name, "myip_port") == 0)
                {
                        n = strtol_safe(value, -1);
                        if(n == -1 || n <=0 || n > 65535)
                        {
                                log_error("Invalid myip port %s", value);
                                ret = -1;
                                break;
                        }

                        cfg->myip.port = n;
                        ++myip_assign_count;
                }
                else if(strcmp(name, "myip_path") == 0)
                {
                        CFG_SET_VALUE(cfg->myip.path, value);
                        ++myip_assign_count;
                }
                else if(strcmp(name, "myip_upint") == 0)
                {
                        n = strtol_safe(value, -1);
                        if(n == -1 || n <=0)
                        {
                                log_error("Invalid myip upint %s", value);
                                ret = -1;
                                break;
                        }

                        cfg->myip.upint = n;
                        ++myip_assign_count;
                }
                else
                {
                        log_error("Invalid option name '%s' (file %s "
                                  "line %d)",
                                  name, filename, linenum);
                        ret = -1;
                        break;
                }

        }

        if(ret == -1 && feof(file))
        {
                log_debug("End of configuration file reached");
                ret = 0;
        }

        if(cfg->wan_cnt_type == wan_cnt_direct
           && cfg->wan_ifname == NULL)
        {
                log_warning("wan ifname isn't defined. Use ppp0 by default");
                cfg->wan_ifname = strdup(CFG_DEFAULT_WANIFNAME);
        }

        if(cfg->wan_cnt_type == wan_cnt_indirect)
        {
                if(cfg->myip.host == NULL
                   || cfg->myip.port == 0
                   || cfg->myip.path == NULL
                   || cfg->myip.upint == 0)
                {
                        log_error("Miss myip definition(s). Check config file");
                        ret = -1;
                }
        }

        if(accountdef_scope)
        {
                log_error("No found closure for account name '%s' service '%s' "
                          "(file %s line %d)",
                          accountcfg->name, accountcfg->service,
                          filename, linenum);
                ret = -1;
        }

        if(ret == -1)
        {
                /* error. need to cleanup */
                CFG_FREE(cfg->wan_ifname);
                CFG_FREE(cfg->myip.host);
                CFG_FREE(cfg->myip.path);

                list_for_each_entry_safe(accountcfg, safe_accountcfg,
                                         &(cfg->accountcfg_list), list)
                {
                        CFG_FREE(accountcfg->name);
                        CFG_FREE(accountcfg->service);
                        CFG_FREE(accountcfg->username);
                        CFG_FREE(accountcfg->passwd);
                        CFG_FREE(accountcfg->name);

                        list_del(&(accountcfg->list));
                        free(accountcfg);
                }

        }

        fclose(file);

	return ret;
}

struct accountcfg * config_account_get(const struct cfg *cfg, const char *name)
{
        struct accountcfg *accountcfg = NULL;

        list_for_each_entry(accountcfg,
                            &(cfg->accountcfg_list), list)
        {
                if(strcmp(accountcfg->name, name) == 0)
                {
                        return accountcfg;
                }
        }

        return NULL;
}

void config_init(struct cfg *cfg)
{
        memset(cfg, 0, sizeof(struct cfg));
        INIT_LIST_HEAD( &(cfg->accountcfg_list) );
}

int config_free(struct cfg *cfg)
{
        struct accountcfg *accountcfg = NULL,
                *safe = NULL;

        CFG_FREE(cfg->wan_ifname);
        CFG_FREE(cfg->myip.host);
        CFG_FREE(cfg->myip.path);
        CFG_FREE(cfg->cfgfile);
        CFG_FREE(cfg->pidfile);

        list_for_each_entry_safe(accountcfg, safe,
                                 &(cfg->accountcfg_list), list)
        {
                CFG_FREE(accountcfg->name);
                CFG_FREE(accountcfg->service);
                CFG_FREE(accountcfg->username);
                CFG_FREE(accountcfg->passwd);
                CFG_FREE(accountcfg->hostname);

                list_del(&(accountcfg->list));
                free(accountcfg);
        }

	return 0;
}

void config_print(struct cfg *cfg)
{
        struct accountcfg *accountcfg = NULL;

        printf("Configuration:\n");
        printf(" cfg file = '%s'\n", cfg->cfgfile);
        printf(" pid file = '%s'\n", cfg->pidfile);
        printf(" daemonize = '%d'\n", cfg->daemonize);
        printf(" wan ifname = '%s'\n", cfg->wan_ifname);
        printf(" wan mode = '%d'\n", cfg->wan_cnt_type);

        list_for_each_entry(accountcfg,
                            &(cfg->accountcfg_list), list)
        {
                printf(" ---- account name '%s' ----\n", accountcfg->name);
                printf("   service = '%s'\n", accountcfg->service);
                printf("   username = '%s'\n", accountcfg->username);
                printf("   password = '%s'\n", accountcfg->passwd);
                printf("   hostname = '%s'\n", accountcfg->hostname);
        }
}

void config_copy(struct cfg *cfgdst, const struct cfg *cfgsrc)
{
        struct accountcfg *actcfg = NULL,
                *safe_actcfg = NULL;

        memcpy(cfgdst, cfgsrc,
               sizeof(struct cfg) - sizeof(struct list_head));
        list_for_each_entry_safe(actcfg, safe_actcfg,
                                 &(cfgsrc->accountcfg_list), list)
        {
                list_move(&(actcfg->list),
                          &(cfgdst->accountcfg_list));
        }
}
